from abc import ABC
from enum import Enum
from functools import lru_cache
from math import sqrt
from operator import itemgetter
from typing import Type, cast, List

import pytextrank  # noqa: F401
import spacy
from collections_extended import RangeMap
from pydantic import BaseModel, Field
from pymultirole_plugins.v1.formatter import FormatterParameters, FormatterBase
from pymultirole_plugins.v1.processor import ProcessorBase, ProcessorParameters
from pymultirole_plugins.v1.schema import Document, Sentence
from spacy.cli.download import download_model, get_compatibility, get_version
from spacy.errors import OLD_MODEL_SHORTCUTS
from spacy.language import Language
from spacy.tokens import Doc
from starlette.responses import Response, PlainTextResponse
from wasabi import msg


# _home = os.path.expanduser('~')
# xdg_cache_home = os.environ.get('XDG_CACHE_HOME') or os.path.join(_home, '.cache')

class TextRankAlgo(str, Enum):
    textrank = 'textrank'
    positionrank = 'positionrank'
    # biasedtextrank = 'biasedtextrank'


class TextRankSummarizerParameters(FormatterParameters):
    as_metadata: str = Field(None, description="""If defined generate the summary as a metadata of the input document,
    if not replace the text of the input document.""")
    lang: str = Field("en",
                      description="Name of the 2-letter language of the documents")
    algo: TextRankAlgo = Field(TextRankAlgo.textrank, description="""The textgraph algorithms to use<br />
    <li>`textrank` TextRank by [mihalcea04textrank](https://web.eecs.umich.edu/~mihalcea/papers/mihalcea.emnlp04.pdf)<br />
    <li>`positionrank` PositionRank by [florescuc17](https://www.aclweb.org/anthology/P17-1102.pdf)<br />""")
    # <li>`biasedtextrank` Biased TextRank by [kazemi-etal-2020-biased](https://www.aclweb.org/anthology/2020.coling-main.144.pdf)<br />""")
    algo: TextRankAlgo = Field(TextRankAlgo.textrank,
                               description="Name of the 2-letter language of the documents")
    num_sentences: float = Field(0.25, description="""Number of sentences of the summary:<br/>
        <li>If float in the range [0.0, 1.0], then consider num_sentences as a percentage of the original number of sentences of the document.""")
    num_phrases: int = Field(10, description="Maximum number of top-ranked phrases to use in the distance vectors.")


class TextRankSummarizerFormatter(FormatterBase, ProcessorBase, ABC):
    """"A graphbased ranking model for text processing. Extractive sentence summarization.
    """

    # cache_dir = os.path.join(xdg_cache_home, 'trankit')
    def _check_sentences(self, nlp, algo: str, document: Document):
        doc: Doc = None
        if not document.sentences:
            doc = nlp(document.text)
        else:
            textrank = nlp.get_pipe(algo)
            parser = nlp.get_pipe("parser")
            with nlp.select_pipes(disable=["parser", algo]):
                doc = nlp(document.text)
                sent_map = RangeMap()
                for i, s in enumerate(document.sentences):
                    if s.end > s.start:
                        sent_map[s.start:s.end] = i
                current = -1
                for token in doc:
                    if token.idx in sent_map:
                        idsent = sent_map[token.idx]
                        if idsent != current:
                            token.is_sent_start = True
                            current = idsent
                if doc.has_annotation("SENT_START"):
                    doc = textrank(parser(doc))
        return doc

    def _summarize(self, document: Document, parameters: FormatterParameters) -> str:
        def int_float(v: float):
            if 0.0 <= v < 1.0:
                return v
            return int(abs(v))

        params: TextRankSummarizerParameters = \
            cast(TextRankSummarizerParameters, parameters)
        nlp = get_nlp(params.lang, params.algo.value)
        doc = self._check_sentences(nlp, params.algo.value, document)
        sent_bounds = [[s.start, s.end, set([])] for s in doc.sents]
        num_sentences = int_float(params.num_sentences)
        limit_sentences = round(len(sent_bounds) * num_sentences) if isinstance(num_sentences, float) else num_sentences
        phrase_id = 0
        unit_vector = []
        for p in doc._.phrases:
            unit_vector.append(p.rank)
            for chunk in p.chunks:
                for sent_start, sent_end, sent_vector in sent_bounds:
                    if chunk.start >= sent_start and chunk.end <= sent_end:
                        sent_vector.add(phrase_id)
                        break
            phrase_id += 1
            if phrase_id == params.num_phrases:
                break

        sum_ranks = sum(unit_vector)

        unit_vector = [rank / sum_ranks for rank in unit_vector]

        sent_rank = {}
        sent_id = 0
        for sent_start, sent_end, sent_vector in sent_bounds:
            sum_sq = 0.0
            for phrase_id in range(len(unit_vector)):
                if phrase_id not in sent_vector:
                    sum_sq += unit_vector[phrase_id] ** 2.0
            sent_rank[sent_id] = sqrt(sum_sq)
            sent_id += 1

        sorted(sent_rank.items(), key=itemgetter(1))

        sent_text = {}
        sent_id = 0
        summary_sentences = []
        for sent in doc.sents:
            sent_text[sent_id] = sent.text.strip()
            sent_id += 1
        num_sent = 0
        for sent_id, rank in sorted(sent_rank.items(), key=itemgetter(1)):
            summary_sentences.append(sent_text[sent_id])
            num_sent += 1
            if num_sent == limit_sentences:
                break
        return summary_sentences

    def format(self, document: Document, parameters: FormatterParameters) \
            -> Response:
        summary_sentences = self._summarize(document, parameters)
        summary = '\n'.join(summary_sentences)
        return PlainTextResponse(summary)

    def process(self, documents: List[Document], parameters: ProcessorParameters) \
            -> List[Document]:
        params: TextRankSummarizerParameters = \
            cast(TextRankSummarizerParameters, parameters)
        for document in documents:
            summary_sentences = self._summarize(document, parameters)
            if params.as_metadata is not None and len(params.as_metadata):
                if not document.metadata:
                    document.metadata = {}
                document.metadata[params.as_metadata] = '\n'.join(summary_sentences)
            else:
                document.text = ""
                document.sentences = []
                for i, summary_sentence in enumerate(summary_sentences):
                    last = i == len(summary_sentences) - 1
                    start = len(document.text)
                    end = start + len(summary_sentence)
                    document.sentences.append(Sentence(start=start, end=end))
                    document.text += summary_sentence + ("\n" if not last else "")
        return documents

    @classmethod
    def get_model(cls) -> Type[BaseModel]:
        return TextRankSummarizerParameters


# Deprecated model shortcuts, only used in errors and warnings
MODEL_SHORTCUTS = {
    "en": "en_core_web_sm", "de": "de_core_news_sm", "es": "es_core_news_sm",
    "pt": "pt_core_news_sm", "fr": "fr_core_news_sm", "it": "it_core_news_sm",
    "nl": "nl_core_news_sm", "el": "el_core_news_sm", "nb": "nb_core_news_sm",
    "lt": "lt_core_news_sm", "xx": "xx_ent_wiki_sm"
}


@lru_cache(maxsize=None)
def get_nlp(lang: str, algo: str, ttl_hash=None):
    del ttl_hash
    model = MODEL_SHORTCUTS.get(lang, lang)
    # model = lang
    try:
        nlp: Language = spacy.load(model)
    except BaseException:
        nlp = load_spacy_model(model)
    nlp.add_pipe(algo, last=True)
    return nlp


def load_spacy_model(model, *pip_args):
    suffix = "-py3-none-any.whl"
    dl_tpl = "{m}-{v}/{m}-{v}{s}#egg={m}=={v}"
    model_name = model
    if model in OLD_MODEL_SHORTCUTS:
        msg.warn(
            f"As of spaCy v3.0, shortcuts like '{model}' are deprecated. Please "
            f"use the full pipeline package name '{OLD_MODEL_SHORTCUTS[model]}' instead."
        )
        model_name = OLD_MODEL_SHORTCUTS[model]
    compatibility = get_compatibility()
    if model_name not in compatibility:
        msg.fail(
            f"No compatible package found for '{model}' (spaCy v{spacy.about.__version__}), fallback to blank model")
        return spacy.blank(model_name)
    else:
        version = get_version(model_name, compatibility)
        download_model(dl_tpl.format(m=model_name, v=version, s=suffix), pip_args)
        msg.good(
            "Download and installation successful",
            f"You can now load the package via spacy.load('{model_name}')",
        )
        # If a model is downloaded and then loaded within the same process, our
        # is_package check currently fails, because pkg_resources.working_set
        # is not refreshed automatically (see #3923). We're trying to work
        # around this here be requiring the package explicitly.
        require_package(model_name)
        return spacy.load(model_name)


def require_package(name):
    try:
        import pkg_resources

        pkg_resources.working_set.require(name)
        return True
    except:  # noqa: E722
        return False



import json
from pathlib import Path

import pytest
from pyformatters_textranksummarizer.textranksummarizer import TextRankSummarizerFormatter, \
    TextRankSummarizerParameters, TextRankAlgo
from pymultirole_plugins.v1.schema import Document, DocumentList
from starlette.responses import Response


@pytest.mark.parametrize("algorithm", [a.value for a in TextRankAlgo])
def test_textranksummarizer_french(algorithm):
    parameters = TextRankSummarizerParameters(lang="fr", algo=algorithm)
    formatter = TextRankSummarizerFormatter()
    doc = Document(text="""Un nuage de fumée juste après l’explosion, le 1er juin 2019.
        Une déflagration dans une importante usine d’explosifs du centre de la Russie a fait au moins 79 blessés samedi 1er juin.
        L’explosion a eu lieu dans l’usine Kristall à Dzerzhinsk, une ville située à environ 400 kilomètres à l’est de Moscou, dans la région de Nijni-Novgorod.
        « Il y a eu une explosion technique dans l’un des ateliers, suivie d’un incendie qui s’est propagé sur une centaine de mètres carrés », a expliqué un porte-parole des services d’urgence.
        Des images circulant sur les réseaux sociaux montraient un énorme nuage de fumée après l’explosion.
        Cinq bâtiments de l’usine et près de 180 bâtiments résidentiels ont été endommagés par l’explosion, selon les autorités municipales. Une enquête pour de potentielles violations des normes de sécurité a été ouverte.
        Fragments de shrapnel Les blessés ont été soignés après avoir été atteints par des fragments issus de l’explosion, a précisé une porte-parole des autorités sanitaires citée par Interfax.
        « Nous parlons de blessures par shrapnel d’une gravité moyenne et modérée », a-t-elle précisé.
        Selon des représentants de Kristall, cinq personnes travaillaient dans la zone où s’est produite l’explosion. Elles ont pu être évacuées en sécurité.
        Les pompiers locaux ont rapporté n’avoir aucune information sur des personnes qui se trouveraient encore dans l’usine.
        """)
    resp: Response = formatter.format(doc, parameters)
    assert resp.status_code == 200
    assert resp.media_type == "text/plain"
    summary = resp.body.decode(resp.charset)

    parameters.as_metadata = "summary"
    docs = formatter.process([doc], parameters)
    summarized: Document = docs[0]
    assert summary == summarized.metadata['summary']

    parameters.as_metadata = None
    docs = formatter.process([doc], parameters)
    summarized: Document = docs[0]
    assert summary == summarized.text


def test_textranksummarizer_french_with_sents():
    testdir = Path(__file__).parent / 'data'
    json_file = testdir / "french.json"
    with json_file.open("r") as fin:
        doc = json.load(fin)
    doc = Document(**doc)
    ori_len = len(doc.text)
    parameters = TextRankSummarizerParameters(lang="fr", num_sentences=2)
    formatter = TextRankSummarizerFormatter()
    docs = formatter.process([doc], parameters)
    assert len(docs[0].text) < ori_len


def test_textranksummarizer_english():
    parameters = TextRankSummarizerParameters(num_sentences=2)
    formatter = TextRankSummarizerFormatter()
    doc = Document(text="""The tower is 324 metres (1,063 ft) tall, about the same height as an 81-storey building, and the tallest structure in Paris.
    Its base is square, measuring 125 metres (410 ft) on each side. During its construction, the Eiffel Tower surpassed the Washington Monument to become the tallest man-made structure in the world, a title it held for 41 years until the Chrysler Building in New York City was finished in 1930.
    It was the first structure to reach a height of 300 metres. Due to the addition of a broadcasting aerial at the top of the tower in 1957, it is now taller than the Chrysler Building by 5.2 metres (17 ft).
    Excluding transmitters, the Eiffel Tower is the second tallest free-standing structure in France after the Millau Viaduct.
        """)
    resp: Response = formatter.format(doc, parameters)
    assert resp.status_code == 200
    assert resp.media_type == "text/plain"
    summary = resp.body.decode(resp.charset)

    parameters.as_metadata = "summary"
    docs = formatter.process([doc], parameters)
    summarized: Document = docs[0]
    assert summary == summarized.metadata['summary']

    parameters.as_metadata = None
    docs = formatter.process([doc], parameters)
    summarized: Document = docs[0]
    assert summary == summarized.text


@pytest.mark.skip(reason="Not a test")
def test_textranksummarizer_kantar():
    testdir = Path(__file__).parent / 'data'
    json_file = testdir / "futilecyclingtxt-documents.json"
    with json_file.open("r") as fin:
        docs = json.load(fin)
    docs = [Document(**doc) for doc in docs]
    parameters = TextRankSummarizerParameters(num_sentences=2)
    formatter = TextRankSummarizerFormatter()
    # rows = []
    for doc in docs:
        resp: Response = formatter.format(doc, parameters)
        assert resp.status_code == 200
        summary = resp.body.decode(resp.charset)
        doc.metadata['summary'] = summary
        # rows.append({
        #     'Title': doc.title,
        #     'Content': doc.text,
        #     'Summary': summary
        # })
    # row_file = testdir / f"{json_file.stem}_{parameters.model.value.replace('/', '_')}.xlsx"
    # df = pd.DataFrame.from_records(rows)
    # df.to_excel(row_file)
    sum_file = testdir / f"{json_file.stem}_textrank.json"
    dl = DocumentList(__root__=docs)
    with sum_file.open("w") as fout:
        print(dl.json(exclude_none=True, exclude_unset=True, indent=2), file=fout)


def test_textranksummarizer_boerhinger():
    parameters = TextRankSummarizerParameters(num_sentences=2)
    formatter = TextRankSummarizerFormatter()
    doc = Document(
        text="Adipose tissue development during early life: novel insights into energy balance from small and large mammals\n\nSince the rediscovery of brown adipose tissue (BAT) in adult human subjects in 2007, there has been a dramatic resurgence in research interest in its role in heat production and energy balance. This has coincided with a reassessment of the origins of BAT and the suggestion that brown preadipocytes could share a common lineage with skeletal myoblasts. In precocial newborns, such as sheep, the onset of non-shivering thermogenesis through activation of the BAT-specific uncoupling protein 1 (UCP1) is essential for effective adaptation to the cold exposure of the extra-uterine environment. This is mediated by a combination of endocrine adaptations which accompany normal parturition at birth and further endocrine stimulation from the mother's milk. Three distinct adipose depots have been identified in all species studied to date. These contain either primarily white, primarily brown or a mix of brown and white adipocytes. The latter tissue type is present, at least, in the fetus and, thereafter, appears to take on the characteristics of white adipose tissue during postnatal development. It is becoming apparent that a range of organ-specific mechanisms can promote UCP1 expression. They include the liver, heart and skeletal muscle, and involve unique endocrine systems that are stimulated by cold exposure and/or exercise. These multiple pathways that promote BAT function vary with age and between species that may determine the potential to be manipulated in early life. Such interventions could modify, or reverse, the normal ontogenic pathway by which BAT disappears after birth, thereby facilitating BAT thermogenesis through the life cycle.")
    resp: Response = formatter.format(doc, parameters)
    assert resp.status_code == 200
    assert resp.media_type == "text/plain"
    summary = resp.body.decode(resp.charset)
    assert "Peter Gough" not in summary
